@@ -70,6 +70,9 @@ int process_rule(KMSI *p_kmsi, XRULE *rp, ITEM *any_index, int usekeys);
 UINT modified_state(UINT state);
 UINT compare_state(ITEM rule_key, ITEM keystroke);
 
+void erase_char_int(KMSI *p_kmsi);
+void queue_item_for_output(KMSI *p_kmsi, ITEM item);
+void process_output_queue(KMSI *p_kmsi);
 void output_item(void *connection, ITEM x);
 void add_to_history(KMSI *p_kmsi,ITEM key);
 void delete_from_history(KMSI *p_kmsi,UINT nchars);
@@ -93,6 +96,8 @@ int kmfl_interpret(KMSI *p_kmsi, UINT key, UINT state)
 	ITEM keysym;
 	int matched;
 
+	p_kmsi->noutput_queue=0;
+	
 	// Test first for modifier key keystrokes and do nothing
 	switch(key) 
 	{
@@ -126,15 +131,21 @@ int kmfl_interpret(KMSI *p_kmsi, UINT key, UINT state)
 
 	// Pass control to the first group for processing, and return if key was matched
 	if((matched=process_group(p_kmsi, p_group1)) > 0) 
+	{
+		process_output_queue(p_kmsi);
 		return 1;
-
+	}
+	
 	// Now try without shift state
 	if ((state & KS_SHIFT) != 0) 
 	{
 		keysym &= ~((unsigned long)KS_SHIFT<<16);
 		p_kmsi->history[0] = keysym;
 		if((matched=process_group(p_kmsi, p_group1)) > 0) 
+		{
+			process_output_queue(p_kmsi);
 			return 1;
+		}
 	}
 
 	/* need some kind of error notification if error value returned */
@@ -144,7 +155,8 @@ int kmfl_interpret(KMSI *p_kmsi, UINT key, UINT state)
 	if(((key & 0xff00) == 0) && ((state & 0xcc) == 0))
 	{
 		add_to_history(p_kmsi,(ITEM)key);
-		output_item(p_kmsi->connection,(ITEM)key);
+		queue_item_for_output(p_kmsi,(ITEM)key);
+		process_output_queue(p_kmsi);
 		return 1;
 	}
 
@@ -154,7 +166,7 @@ int kmfl_interpret(KMSI *p_kmsi, UINT key, UINT state)
 
 	case 0xff08:		// backspace - erase last character from history
 		delete_from_history(p_kmsi,1);
-		erase_char(p_kmsi->connection);
+		erase_char_int(p_kmsi);
 		return 1;
 	case 0xff09:		// tab - clear history, let app handle key
 	case 0xff0d:		// return - clear history, let app handle key
@@ -246,7 +258,9 @@ int match_rule(KMSI *p_kmsi, XRULE *rp, ITEM *any_index, int usekeys)
 	
 	for(m=0; m<rp->ilen; pr++,ph--,m++) 
 	{
-		switch(ITEM_TYPE(*pr)) 
+		unsigned char item_type;
+		item_type=ITEM_TYPE(*pr);
+		switch(item_type) 
 		{
 		case ITEM_CHAR:
 			if(*pr != *ph) return 0;
@@ -262,6 +276,7 @@ int match_rule(KMSI *p_kmsi, XRULE *rp, ITEM *any_index, int usekeys)
 			break;				// matched - continue matching string
 
 		case ITEM_ANY:	// will need to allow for matching keysyms in any()
+		case ITEM_NOTANY: 
 			ps = store_content(p_kmsi,STORE_NUMBER(*pr));
 			nmax = store_length(p_kmsi,STORE_NUMBER(*pr));
 			if(m == rp->ilen-1) mask = 0xffffff; else mask = 0xffffffff;
@@ -273,7 +288,11 @@ int match_rule(KMSI *p_kmsi, XRULE *rp, ITEM *any_index, int usekeys)
 					break;
 				}
 			}
-			if(n == nmax) return 0;		// no match
+			if (item_type == ITEM_ANY) {
+				if(n == nmax) return 0;		// no match
+			} else {
+				if(n != nmax) return 0;		// no match
+			}
 			break;				// matched - continue matching string
 
 		case ITEM_INDEX:		// indexes start from 1, not 0
@@ -353,8 +372,10 @@ int process_rule(KMSI *p_kmsi, XRULE *rp, ITEM *any_index, int usekeys)
 		case ITEM_NOMATCH:
 			break;
 		default:
-			if(itp != ITEM_DEADKEY) erase_char(p_kmsi->connection);	
-			for(i=1; i<p_kmsi->nhistory; i++) p_kmsi->history[i] = p_kmsi->history[i+1];
+			if(ITEM_TYPE(p_kmsi->history[1]) != ITEM_DEADKEY) 
+				erase_char_int(p_kmsi);	
+			for(i=1; i<p_kmsi->nhistory; i++) 
+				p_kmsi->history[i] = p_kmsi->history[i+1];
 			p_kmsi->nhistory--;
 			break;
 		}
@@ -481,7 +502,7 @@ int process_rule(KMSI *p_kmsi, XRULE *rp, ITEM *any_index, int usekeys)
                 } 
                 else
                 {
-                	output_item(p_kmsi->connection,*p);
+                	queue_item_for_output(p_kmsi,*p);
                 	add_to_history(p_kmsi,*p);
                	}
             }
@@ -494,6 +515,33 @@ int process_rule(KMSI *p_kmsi, XRULE *rp, ITEM *any_index, int usekeys)
 	return retCode;	// Return 1 (or 2) to indicate that the keystroke has been matched and processed
 }
 
+// Check to see if there are deadkeys in the current history
+int deadkey_in_history(KMSI *p_kmsi)
+{
+	ITEM * pitem =p_kmsi->history+1;
+	UINT nitems= p_kmsi->nhistory;
+	UINT iitem;
+
+	for (iitem=0; iitem < nitems; iitem++, pitem++) {
+		if(ITEM_TYPE(*pitem) == ITEM_DEADKEY) {
+			return 1; 
+		}
+	}
+
+	return 0;
+}
+
+// Sets the history to the surrounding context 
+void set_history(KMSI *p_kmsi, ITEM * items, UINT nitems)
+{
+
+	if (nitems > MAX_HISTORY)
+		nitems = MAX_HISTORY;
+
+	memcpy(p_kmsi->history+1, items, nitems * sizeof(ITEM));
+	p_kmsi->nhistory=nitems;
+}
+
 // Add a character item (or deadkey) to the start of the history stack (to item 1)
 void add_to_history(KMSI *p_kmsi,ITEM item) 
 {
@@ -514,9 +562,11 @@ void delete_from_history(KMSI *p_kmsi,UINT nitems)
 {
 	UINT nleft;
 
-	if(p_kmsi->nhistory > MAX_HISTORY) p_kmsi->nhistory = MAX_HISTORY;
+	if(p_kmsi->nhistory > MAX_HISTORY) 
+		p_kmsi->nhistory = MAX_HISTORY;
 
-	if(nitems > p_kmsi->nhistory) nitems = p_kmsi->nhistory;
+	if(nitems > p_kmsi->nhistory) 
+		nitems = p_kmsi->nhistory;
 	nleft = p_kmsi->nhistory - nitems;
 	if(nleft > 0 && nitems > 0)
 	{
@@ -532,13 +582,69 @@ void clear_history(KMSI *p_kmsi)
 	p_kmsi->nhistory = 0;
 }
 
+// a single key event we queue items for output. If the keyboard needs to delete a 
+// character, this queue is checked first and characters are deleted first from
+// here. If this queue is empty, then characters are delete from the application
+// This avoids cases such <BS><BS><C1><BS><C2><C3> where <C1> <C2> and <C3> are
+// characters and <BS> is a backspace. What the application will see is 
+// OB<BS><BS><C2><C3>
+void queue_item_for_output(KMSI *p_kmsi, ITEM item)
+{
+	if (p_kmsi->noutput_queue < MAX_OUTPUT) 
+	{
+		p_kmsi->output_queue[p_kmsi->noutput_queue]= item;
+		(p_kmsi->noutput_queue)++;
+	} else {
+		ERRMSG("Exceeded maximum length of output allowed from any one key event.\n");
+	}
+}
+
+void process_output_queue(KMSI *p_kmsi)
+{
+	int i;
+	
+	UTF32 utfin[2]={0};
+	UTF32 *pin;
+	UTF8 utfout[MAX_OUTPUT*4+1]={0};
+	UTF8 *pout;
+	size_t result;
+	
+	pout = &utfout[0];
+	for (i=0; i < p_kmsi->noutput_queue; i++) {
+#if 0
+		output_item(p_kmsi->connection, p_kmsi->output_queue[i]);
+#else
+		pin = &utfin[0]; 
+		utfin[0] = p_kmsi->output_queue[i]; 
+		result = IConvertUTF32toUTF8((const UTF32 **)&pin,utfin+1,&pout,utfout+MAX_OUTPUT*4);
+		if (result == (size_t)-1) {
+			ERRMSG("Exceeded maximum length of output allowed from any one key event.\n");
+			return;
+		}
+	}
+	*pout = 0;
+	output_string(p_kmsi->connection, (char *)utfout);
+#endif
+}
+
+void erase_char_int(KMSI *p_kmsi)
+{
+	if (p_kmsi->noutput_queue > 0)
+		(p_kmsi->noutput_queue)--;
+	else
+		erase_char(p_kmsi->connection);
+}
+
+// Because some apps cannot handle a mixture of erases and commits when processing
 // Output a Unicode character (as a multi-byte string)
 void output_item(void *connection, ITEM x)
 {
 	UTF32 utfin[2]={0}, *pin;
 	UTF8 utfout[16]={0}, *pout;
 	size_t result;
-	utfin[0] = x; pin = &utfin[0]; pout = &utfout[0];
+	utfin[0] = x; 
+	pin = &utfin[0]; 
+	pout = &utfout[0];
 
 	result = IConvertUTF32toUTF8((const UTF32 **)&pin,utfin+1,&pout,utfout+15);
 	
